from uuid import UUID
from ..savable import Savable
from ..resource import Resource
from ...errors import ResourceException
from typing import Any, Dict, Callable
import jsonpickle
from ..db.constraint import Constraint
from ...errors import EpureError
# from .elist_metacls import ElistMetacls
from .elist_metacls import ECollectionMetacls
from ..node_promise import NodePromise, ElistPromise

class Node(Savable):

    node_id:object
    __exclude__:list
    __promises_dict__:Dict[str, Any]

    def __init__(self, node_id:object=None, resource:Resource=None) -> None:
        if node_id != None:
            self.node_id = node_id
        super().__init__(resource)

    # def __getattribute__(self, name: str) -> Any:
    #     from ..node_promise import NodePromise
    #     val = super(Node, self).__getattribute__(name)
    #     if isinstance(val, NodePromise):
    #         setattr(self, name, val.get())
    #     return super().__getattribute__(name)
    
    def __getattr__(self, name:str):
        if '__promises_dict__' in self.__dict__ and name in self.__promises_dict__:
            res = self.__promises_dict__[name].get()
            setattr(self, name, res)
            self.__promises_dict__.pop(name)
            return res
        else:
            raise AttributeError(f"'{type(self)}' object has no attribute '{name}'")
        
    def load(self):
        if '__promises_dict__' in self.__dict__ and self.__promises_dict__:
            for name in list(self.__promises_dict__):
                self.__getattr__(name)
            delattr(self, "__promises_dict__")

    @classmethod
    def from_dict(_cls, _dict:Dict[str, Any])->object:
        raise NotImplementedError

    def save(self, asynch:bool=False):
        if not hasattr(self, 'resource'):
            raise ResourceException('Unable save Savable, resource not found')
        resource = self.resource
        
        
        if hasattr(self, 'node_id') and self.node_id:
            res = resource.update(self, asynch)
            return res
        else:
            res = resource.create(self, asynch)
            return res


class TableNode(Node):
    db:Resource
    node_id: UUID
    resource:Savable

    def __init__(self, node_id:object=None, resource:Resource=None) -> None:
        if node_id != None and isinstance(node_id, str):
            node_id = UUID(node_id)
        super().__init__(node_id, resource)

    @property
    def table(self) -> Savable:
        return self.resource

    @classmethod
    def from_dict_deep(_cls, _dict:Dict[str, Any])->object:

        # instance = _cls.__call__()
        instance = _cls()
        
        for field_name, val in _dict.items():
            if field_name in instance.annotations:
                _type = instance.annotations[field_name]
                if isinstance(_type, Savable):
                    val = _type(val)
                setattr(instance, field_name, val)


        return instance
    
    @classmethod
    def from_dict(_cls, _dict:Dict[str, Any])->object:
        from .elist import ECollectionMetacls

        # instance = _cls.__call__()
        instance = _cls()
        instance.__promises_dict__ = {}

        for field_name, val in _dict.items():

            if not field_name in instance.annotations:
                continue

            cls_attr_type = instance.annotations[field_name]
            
            val_type_match_cls_attr_type = isinstance(val, cls_attr_type)

            if field_name == "node_id" and not isinstance(val, UUID):
                val = UUID(val)
                val_type_match_cls_attr_type = True

            if isinstance(cls_attr_type, ECollectionMetacls) and not val_type_match_cls_attr_type: # check if attr is elist
                try: 
                    collection_node_id = UUID(val)
                    promise = ElistPromise(cls_attr_type.list_epure.resource, collection_node_id, cls_attr_type)
                    instance.__promises_dict__[field_name] = promise
                    continue
                except(Exception):
                    if cls_attr_type.py_type in (bytes, bytearray)\
                    and not type(val[0]) in (bytes, bytearray):
                        for i, item in enumerate(val):
                            val[i] = item.encode()
                    val = cls_attr_type(val)

                val_type_match_cls_attr_type = True

            elif issubclass(cls_attr_type, Savable) and not val_type_match_cls_attr_type: # check if attr is epure 
                try: 
                    node_id = UUID(val)
                    promise = NodePromise(cls_attr_type.resource, node_id)
                    instance.__promises_dict__[field_name] = promise
                    continue
                except(Exception):
                    val = cls_attr_type(val)

                val_type_match_cls_attr_type = True
            
            if val is None:
                val_type_match_cls_attr_type = True

            if isinstance(val, str) and cls_attr_type in (bytes, bytearray): # bytearray is yet to be tested!
                val = cls_attr_type(val.encode())
                val_type_match_cls_attr_type = True

            if val_type_match_cls_attr_type:
                setattr(instance, field_name, val)
            else:
                raise TypeError(f'Value for field "{field_name}" with value "{val}" of type "{type(val)}" does not match expected attr type of class '\
                                 f'"{_cls}" with value name "{field_name}" and type of attr "{cls_attr_type}"')


        return instance
    
    # def to_json(self) -> str:
    #     from ..db.table import NodePromise
    #     res = {}
    #     for field_name, field_type in self.annotations.items():
    #         if isinstance(field_type, Constraint):
    #             field_type = field_type.py_type
                
    #         if self.is_excluded(field_name, field_type):
    #             continue
    #         if field_name not in self.table.header:
    #             continue
    #         if not hasattr(self, field_name):
    #             continue

    #         field_val = getattr(self, field_name, None)

    #         if isinstance(field_val, NodePromise):
    #             field_val = getattr(field_val, 'node_id', None)
    #             # field_val = getattr(field_val, 'none2', None)

    #         #working for db:
    #         if field_val and isinstance(field_type, Savable)\
    #         and not isinstance(field_val, UUID):
    #             field_val = getattr(self, field_name, None)
    #             field_type = field_val.annotations['node_id']
    #             field_val = field_val.save(True).node_id

    #         if isinstance(field_val, UUID):
    #             field_val = str(field_val)

    #         res[field_name] = field_val

    #     jsonpickle.set_preferred_backend('json')
    #     jsonpickle.set_encoder_options('json', ensure_ascii=False)

    #     res = jsonpickle.encode(res)

    #     return res

    def _serialize_field_val_to_dict(self, field_val, field_type=None, field_name:str=None, rec_depth:int = None, *args):
        # return super()._serialize_field_val(field_type)
        lambda_func = args[0][0]

        if isinstance(field_val, Savable) and not isinstance(type(field_val), ECollectionMetacls)\
        and lambda_func(field_name, field_val, self, rec_depth, args):
            field_val = field_val.to_dict(rec_depth+1, lambda_func)

        elif (isinstance(type(field_val), ECollectionMetacls) or type(field_val) in (list,)) and isinstance(field_val[0], Savable)\
        and lambda_func(field_name, field_val, self, rec_depth, args):
            field_val = [field_val[i].to_dict(rec_depth+1, lambda_func) for i in range(len(field_val))]

        elif isinstance(type(field_val), ECollectionMetacls) and not isinstance(field_val[0], Savable)\
        and lambda_func(field_name, field_val, self, rec_depth, args):
            field_val = [field_val[i] for i in range(len(field_val))]

        elif isinstance(field_val, Savable):
            field_val = field_val.save(True).node_id

        if isinstance(field_val, UUID):
            field_val = str(field_val)

        return field_val


    def to_dict(self, rec_depth=0, lambda_func:Callable = lambda field_name, field_value, parent_value, rec_depth, args: 
                rec_depth < 1 or isinstance(type(parent_value), ECollectionMetacls)) -> Dict[str, Any]:
        from .elist import ECollectionMetacls

        self.load()

        # _dict = self.__dict__.copy()
        _dict = self._serialize(self, self._serialize_field_val_to_dict, rec_depth, lambda_func)

        # if "__promises_dict__" in _dict and _dict['__promises_dict__']:
        #     for name, value in _dict['__promises_dict__'].items():
        #         if isinstance(value, ElistPromise):
        #             # value = value.get().read()
        #             value = value.get()
        #         else:
        #             value = getattr(value, 'node_id', None)
        #         _dict[name] = value

        # _dict.pop("__promises_dict__", None)

        # for field_name, field_val in _dict.items():
        #     if isinstance(field_val, Constraint):
        #         field_val = field_val.py_type

        #     # if isinstance(type(field_val), ElistMetacls) and not issubclass(field_val.py_type, Savable):
        #     if isinstance(type(field_val), ECollectionMetacls):
        #         field_val = [field_val[i] for i in range(len(field_val))]

        #     elif isinstance(field_val, Savable) and lambda_func(field_name, field_val, self, rec_depth):
        #         # field_val = getattr(self, field_name, None)
        #         # field_val = field_val.save(True).node_id
        #         field_val = field_val.to_dict(rec_depth+1, lambda_func)

        #     if isinstance(field_val, UUID):
        #         field_val = str(field_val)

        #     _dict[field_name] = field_val
 
        return _dict

    def to_json(self, encoder:Callable=jsonpickle.encode) -> Dict[str, Any]:

        _dict = self.to_dict()
        
        return encoder(_dict)

    __exclude__:list = Node.__exclude__ + ['table', 'db']

class EsetTableNode(TableNode):

    def __hash__(self) -> int:
        val = self.value
        # if hasattr(val, "node_id") and val.node_id:
        #     return hash(val.node_id)
        
        return hash(id(val))


from types import NoneType
from typing import Any, Type, Dict
from uuid import UUID

# class EcollectionMetacls(type):
#     py_type:type = NoneType
#     __origin__:type = NoneType
#     ecollections:Dict[str,type] = {}

#     def __getitem__(self:Type, param:Any):
#         from ...epure import Epure, epure
#         name = f"{param.__name__}__list"
#         if name in self.elists:
#             return self.elists[name]
        
#         cls_dict = dict(self.__dict__)
#         cls_dict.pop('__dict__', None)
#         res = self.__class__(self.__name__, self.__bases__, cls_dict)
#         # res.__origin__ = self
#         res.py_type = param
#         # res.list_epure = Epure.EDb.get_epure_by_table_name(name)
#         # if res.list_epure is None:
#         obj = type(name, (object,), {})
#         obj.__annotations__ = {"elist_node_id":UUID, "value_order":int, "value":param}
#         # res.list_epure = epure(saver=EsetTableNode)(obj)
#         res.list_epure = epure()(obj)
#         self.elists[name] = res

#         # if issubclass(res.py_type, Constraint):
#         #     # ( hasattr(res.py_type, '__origin__') and  issubclass(res.py_type.__origin__, Constraint)):
#         #     raise EpureError('Constraint parameter cant be Constraint')
#         return res

class ECollectionMetacls(type):
    py_type:type = NoneType
    __origin__:type = NoneType
    ecollections:Dict[str,type] = {}

    def __getitem__(self:Type, param:Any):
        from ...epure import epure
        from .node import EsetTableNode
        from .elist import Elist

        if issubclass(self, Elist):
            name = f"{param.__name__}__list"
        else:
            name = f"{param.__name__}__set"

        if name in self.ecollections:
            return self.ecollections[name]
        
        cls_dict = dict(self.__dict__)
        cls_dict.pop('__dict__', None)
        res = self.__class__(self.__name__, self.__bases__, cls_dict)
        # res.__origin__ = self
        res.py_type = param
        # res.collection_epure = Epure.EDb.get_epure_by_table_name(name)
        # if res.collection_epure is None:
        obj = type(name, (object,), {})
        # res.collection_epure = epure(saver=EsetTableNode)(obj)
        
        if issubclass(self, Elist):
            obj.__annotations__ = {"collection_node_id":UUID, "value_order":int, "value":param}
            res.collection_epure = epure()(obj)
        else:
            obj.__annotations__ = {"collection_node_id":UUID, "value":param}
            res.collection_epure = epure(saver=EsetTableNode)(obj)
        
        self.ecollections[name] = res

        # if issubclass(res.py_type, Constraint):
        #     # ( hasattr(res.py_type, '__origin__') and  issubclass(res.py_type.__origin__, Constraint)):
        #     raise EpureError('Constraint parameter cant be Constraint')
        return res



from __future__ import annotations
from collections.abc import Iterable
from .node import TableNode
from typing import List, Any, Type, Generic, Set, Dict
from uuid import UUID, uuid4
from ...epure import Epure, epure
from types import NoneType
# from .elist_metacls import ElistMetacls
from .elist_metacls import ECollectionMetacls

class Elist(TableNode, List, metaclass=ECollectionMetacls):
# class Elist(TableNode, List):
    entries:List
    deleted_entries:List
    py_type:type = NoneType
    collection_epure:Epure = None
    node_id:UUID

    def __init__(self, _list:List) -> None:
        self.entries = []
        self.deleted_entries = []
        if isinstance(_list[0], self.collection_epure):
            self.node_id = _list[0].collection_node_id
            _list.sort(key= lambda x: x.value_order)
            self.entries = _list
        else:
            for val in _list:
                self.append(val)

    def save(self, asynch:bool=False) -> UUID:
            
        if not hasattr(self,"node_id") or not self.node_id:
            self.node_id = uuid4()

        val_len = self.entries.__len__()
        for i in range(val_len):
            item = self.entries[i]
            if not isinstance(item.value, self.py_type):
                raise TypeError(f"value '{item.value}' of type '{type(item.value)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
            
            item.collection_node_id = self.node_id
            item.value_order = i

            # if hasattr(val, "__deleted__") and val.__deleted__:
            #     self.resource.delete(val)
            if i != val_len-1:
                item.save(asynch=True)
            else:
                item.save(asynch=asynch)

        for i in range(self.deleted_entries.__len__()):
            item = self.deleted_entries[i]
            self.collection_epure.resource.delete(item.node_id)
        
        self.deleted_entries = []

        return self

    def __setitem__(self, index, item) -> None:
        if not isinstance(item, self.py_type):
                raise TypeError(f"value '{item}' of type '{type(item)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        self.entries[index].value = item
        # res = self.collection_epure()
        # res.value_order = index
        # res.value = item
        # self.entries.__setitem__(index, res)

    # def __delitem__(self, key)

    def insert(self, index, item) -> None:
        if not isinstance(item, self.py_type):
                raise TypeError(f"value '{item}' of type '{type(item)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        res = self.collection_epure()
        res.value_order = index
        res.value = item
        self.entries.insert(index, res)

    def append(self, item) -> None:
        if not isinstance(item, self.py_type):
                raise TypeError(f"value '{item}' of type '{type(item)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        res = self.collection_epure()
        res.value_order = len(self.entries)
        res.value = item
        self.entries.append(res)

    def count(self) -> int:
        return self.entries.count()
    
    def copy(self) -> list:
        return self.entries.copy()
    
    def clear(self) -> None:
        entries = self.entries.copy()
        self.entries.clear()
        self.deleted_entries.extend(entries)

    def index(self, value:Any, *args) -> int:
        return self.entries.index(value, *args)

    def extend(self, __iterable: Iterable) -> None:
        self.entries.extend(__iterable)
    
    def pop(self, index = -1) -> Any:
        res = self.entries.pop(index)
        self.deleted_entries.append(res)
        return res
    
    def remove(self, __value: Any) -> None:
        for x in self.entries:
            if x.value == __value:
                __value = x
                break
        self.entries.remove(__value)
        self.deleted_entries.append(__value)
    
    def reverse(self) -> None:
        self.entries.reverse()
    
    def sort(self, *, key=None, reverse=False) -> None:
        self.entries.sort(key=key, reverse=reverse)

    def __getitem__(self: Type, _param: Any):
        return self.entries[_param].value
    
    def __repr__(self):
        return str(self.entries)
    
    # def __eq__(self, __value: Elist|object) -> bool:
    #     if issubclass(__value,Elist):
    #         return self.entries.__eq__(__value.entries)
    #     return self.entries.__eq__(__value)

    def read(self, *args, **kwargs):
        if not isinstance(self.py_type, Epure):
            return

        node_id_dict:dict = {}
        
        for val in self.entries:
            if hasattr(val, '__promises_dict__') and 'value' in val.__promises_dict__:
                node_id_dict[val.__promises_dict__['value'].node_id] = val

        if len(node_id_dict) == 0:
            return

        res = self.py_type.resource.read(lambda tp, dp: tp.node_id >= list(node_id_dict.keys()))
        # res = self.py_type.resource.read(lambda tp, dp: tp.node_id >= list(node_id_dict.values()))

        for val in res:
            node_id_dict[val.node_id].value = val

    def __len__(self) -> int:
        return self.entries.__len__()
    
    @property
    def ids(self):
        res = []
        for item in self.entries:
            # if hasattr(item, "__promises_dict__") and\
            # "value" in item.__promises_dict__:
            #     id = item.__promises_dict__['value'].node_id
            # else:
            id = item.node_id

            res.append(str(id))

        return res

class Eset(set, TableNode, metaclass=ECollectionMetacls):
    # entries:Set
    deleted_entries:Set
    py_type:type = NoneType
    collection_epure:Epure = None
    node_id:UUID

    def __init__(self, _set:Set) -> None:
        super(set, self).__init__()
        self.deleted_entries = []
        if isinstance(_set[0], self.collection_epure):
                self.node_id = _set[0].collection_node_id
                super(self.__class__, self).update(_set)
        else:
            for item in _set:
                self.add(item)

    def add(self, item: Any) -> None:

        if not isinstance(item, self.py_type):
                raise TypeError(f"value '{item}' of type '{type(item)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
        
        # super(Eset, self).add(item)
        # self.add(item)
        res = self.collection_epure()
        res.value = item
        for val in self:
            if item is val.value:
                raise ValueError(f"value {item} is already present in Eset")
        super(self.__class__, self).add(res)

    def save(self, asynch:bool=False) -> UUID:
            
        if not hasattr(self, "node_id") or not self.node_id:
            self.node_id = uuid4()

        val_len = self.__len__()
        self_list = list(self)
        for i in range(val_len):
            item = self_list[i]
            if not isinstance(item.value, self.py_type):
                raise TypeError(f"value '{item.value}' of type '{type(item.value)}' is not same " 
                                f"type as Elist '{self.collection_epure.resource.full_name}' type of '{self.py_type}'")
            
            item.collection_node_id = self.node_id

            # if hasattr(val, "__deleted__") and val.__deleted__:
            #     self.resource.delete(val)
            if i != val_len-1:
                item.save(asynch=True)
            else:
                item.save(asynch=asynch)

        for i in range(self.deleted_entries.__len__()):
            item = self.deleted_entries[i]
            self.collection_epure.resource.delete(item.node_id)
        
        self.deleted_entries = []

        return self
    
    def remove(self, ___element: Any) -> None:
        for x in self:
            if x.value == ___element:
                ___element = x
                break
        self.deleted_entries.append(___element)
        return super(self.__class__, self).remove(___element)
    
    def discard(self, el: Any) -> None:
        for x in self:
            if x.value == el:
                el = x
                break
        self.deleted_entries.append(el)
        return super(self.__class__, self).discard(el)
    
    def clear(self) -> None:
        entries = self.copy()
        self.deleted_entries.extend(entries)
        return super(self.__class__, self).clear()
    
    @property
    def ids(self):
        res = []
        for item in list(self):
            # if hasattr(item, "__promises_dict__") and\
            # "value" in item.__promises_dict__:
            #     id = item.__promises_dict__['value'].node_id
            # else:
            id = item.node_id

            res.append(str(id))

        return res
    
    def __contains__(self, __o: object) -> bool:
        
        for x in self:
            if x.value == __o:
                return True

        # return super().__contains__(__o)


from __future__ import annotations
from types import NoneType
from .resource import Resource
from typing import Any, Dict, get_type_hints, Callable
from .db.constraint import Constraint

class Savable(Resource):    
    
    resource: Resource
    is_saved:bool=False
    _annotations: Dict[str,Any]
    #append to it, if u want exclude
    __exclude__:list = ['resource', 'is_saved', '_annotations', 'cache_queue']
    

    def __init__(self, resource:Resource=None) -> None:
        if resource != None:
            self.resource = resource
        super().__init__()

    @property
    def annotations(self) -> Dict[str,Any]:
        if not hasattr(self, '_annotations'):
            if isinstance(self, type):                
                self._annotations = get_type_hints(self)
            else:
                self.__class__._annotations = get_type_hints(self.__class__)
        return self._annotations

    def save(self, asynch:bool=False):
        pass

    # def to_json(self) -> str:
    #     raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        raise NotImplementedError

    @classmethod
    def is_excluded(self, node, atr_name:str, type_hint:Any='') -> bool:
        cls = type(node)
        if isinstance(node, type):
            cls = node

        if hasattr(cls, '__exclude__') and atr_name in cls.__exclude__:
            return True
        if atr_name[:2] == "__" and atr_name[-2:] == "__":
            return True
        if atr_name in ("prepared_resource"):
            return True
        # if type_hint in (NoneType, None):
        #     return True
        return False

    def execute(self, script: str = '') -> object:
        return self.resource.execute(script)
    
    def _serialize(self, node: Savable, serializer:Callable, rec_depth:int=None, *args) -> Dict[str, str]:
        res = {}
        for field_name, field_type in node.annotations.items():
            if isinstance(field_type, Constraint):
                field_type = field_type.py_type
                
            # if node.is_excluded(field_name, field_type):
            #     continue
            if self.is_excluded(node, field_name, field_type):
                continue
            # if field_name not in self.header:
            #     continue
            if not hasattr(node, field_name):
                continue

            field_val = getattr(node, field_name, None)

            field_val = serializer(field_val, field_type, field_name, rec_depth, args)
            
            # field_val = self._serialize_field_val(field_val, field_type)

            res[field_name] = field_val
        
        return res
    
    # def _serialize_field_val(field_val, field_type=None):
    #     raise NotImplementedError